@@ -5,12 +5,14 @@ import os
 import sys
 import threading
 from collections import defaultdict
-from concurrent.futures import ThreadPoolExecutor
+from concurrent.futures import Future, ThreadPoolExecutor
 from contextlib import asynccontextmanager, suppress
 from dataclasses import dataclass, field, replace
 from functools import wraps
 from os import PathLike
+from queue import Queue
 from typing import (
+    TYPE_CHECKING,
     Any,
     Callable,
     ClassVar,
@@ -26,25 +28,31 @@ from typing import (
 
 import cloudpickle
 import grpc
-import isolate
 import tblib
 import uvicorn
 import yaml
 from fastapi import FastAPI
 from fastapi import __version__ as fastapi_version
-from isolate.backends.common import active_python
-from isolate.backends.settings import DEFAULT_SETTINGS
-from isolate.connections import PythonIPC
+from fastapi.encoders import jsonable_encoder
+from fastapi.exceptions import RequestValidationError
 from packaging.requirements import Requirement
 from packaging.utils import canonicalize_name
 from pydantic import __version__ as pydantic_version
 from typing_extensions import Concatenate, ParamSpec
 
 import fal.flags as flags
-from fal._serialization import include_modules_from, patch_pickle
-from fal.exceptions import FalServerlessException
+from fal._serialization import include_module, include_modules_from, patch_pickle
+from fal.container import ContainerImage
+from fal.exceptions import (
+    AppException,
+    CUDAOutOfMemoryException,
+    FalServerlessException,
+    FieldException,
+)
+from fal.exceptions._cuda import _is_cuda_oom_exception
 from fal.logging.isolate import IsolateLogPrinter
 from fal.sdk import (
+    FAL_SERVERLESS_DEFAULT_CONCURRENCY_BUFFER,
     FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
     FAL_SERVERLESS_DEFAULT_MAX_MULTIPLEXING,
     FAL_SERVERLESS_DEFAULT_MIN_CONCURRENCY,
@@ -57,6 +65,9 @@ from fal.sdk import (
     get_default_credentials,
 )
 
+if TYPE_CHECKING:
+    from isolate.backends import BaseEnvironment
+
 ArgsT = ParamSpec("ArgsT")
 ReturnT = TypeVar("ReturnT", covariant=True)  # noqa: PLC0105
 
@@ -68,6 +79,9 @@ SERVE_REQUIREMENTS = [
     f"pydantic=={pydantic_version}",
     "uvicorn",
     "starlette_exporter",
+    "structlog",
+    "tomli",
+    "tomli-w",
 ]
 
 
@@ -82,8 +96,32 @@ class InternalFalServerlessError(FalServerlessException):
 
 
 @dataclass
-class FalMissingDependencyError(FalServerlessError):
-    ...
+class FalMissingDependencyError(FalServerlessError): ...
+
+
+@dataclass
+class SpawnInfo:
+    future: Future | None = None
+    logs: Queue = field(default_factory=Queue)
+    _url_ready: threading.Event = field(default_factory=threading.Event)
+    _url: str | None = None
+    stream: Any = None
+
+    @property
+    def return_value(self):
+        if self.future is None:
+            raise ValueError
+        return self.future.result()
+
+    @property
+    def url(self):
+        self._url_ready.wait()
+        return self._url
+
+    @url.setter
+    def url(self, value):
+        self._url_ready.set()
+        self._url = value
 
 
 @dataclass
@@ -105,6 +143,8 @@ class Host(Generic[ArgsT, ReturnT]):
             # Conda environment definition should be parsed before sending to serverless
             with open(value) as f:
                 return "env_dict", yaml.safe_load(f)
+        elif key == "image" and isinstance(value, ContainerImage):
+            return "image", value.to_dict()
         else:
             return key, value
 
@@ -128,17 +168,6 @@ class Host(Generic[ArgsT, ReturnT]):
 
         return options
 
-    def register(
-        self,
-        func: Callable[ArgsT, ReturnT],
-        options: Options,
-        application_name: str | None = None,
-        application_auth_mode: Literal["public", "shared", "private"] | None = None,
-        metadata: dict[str, Any] | None = None,
-    ) -> str | None:
-        """Register the given function on the host for API call execution."""
-        raise NotImplementedError
-
     def run(
         self,
         func: Callable[ArgsT, ReturnT],
@@ -149,6 +178,15 @@ class Host(Generic[ArgsT, ReturnT]):
         """Run the given function in the isolated environment."""
         raise NotImplementedError
 
+    def spawn(
+        self,
+        func: Callable[ArgsT, ReturnT],
+        options: Options,
+        args: tuple[Any, ...],
+        kwargs: dict[str, Any],
+    ) -> SpawnInfo:
+        raise NotImplementedError
+
 
 def cached(func: Callable[ArgsT, ReturnT]) -> Callable[ArgsT, ReturnT]:
     """Cache the result of the given function in-memory."""
@@ -171,9 +209,10 @@ def cached(func: Callable[ArgsT, ReturnT]) -> Callable[ArgsT, ReturnT]:
     ) -> ReturnT:
         from functools import lru_cache
 
-        # HACK: Using the isolate module as a global cache.
         import isolate
 
+        # HACK: Using the isolate module as a global cache.
+
         if not hasattr(isolate, "__cached_functions__"):
             isolate.__cached_functions__ = {}
 
@@ -218,17 +257,23 @@ def _prepare_partial_func(
     return wrapper
 
 
-@dataclass
-class LocalHost(Host):
-    # The environment which provides the default set of
-    # packages for isolate agent to run.
-    _AGENT_ENVIRONMENT = isolate.prepare_environment(
+def _prepare_environment() -> BaseEnvironment:
+    import isolate
+
+    return isolate.prepare_environment(
         "virtualenv",
         requirements=[
             f"cloudpickle=={cloudpickle.__version__}",
             f"tblib=={tblib.__version__}",
         ],
     )
+
+
+@dataclass
+class LocalHost(Host):
+    # The environment which provides the default set of
+    # packages for isolate agent to run.
+    _AGENT_ENVIRONMENT: BaseEnvironment = field(default_factory=_prepare_environment)
     _log_printer = IsolateLogPrinter(debug=flags.DEBUG)
 
     def run(
@@ -238,6 +283,10 @@ class LocalHost(Host):
         args: tuple[Any, ...],
         kwargs: dict[str, Any],
     ) -> ReturnT:
+        import isolate
+        from isolate.backends.settings import DEFAULT_SETTINGS
+        from isolate.connections import PythonIPC
+
         settings = replace(
             DEFAULT_SETTINGS,
             serialization_method="cloudpickle",
@@ -270,25 +319,26 @@ def _handle_grpc_error():
             try:
                 return fn(*args, **kwargs)
             except grpc.RpcError as e:
+                msg = e.details() or str(e)
                 if e.code() == grpc.StatusCode.UNAVAILABLE:
                     raise FalServerlessError(
                         "Could not reach fal host. "
                         "This is most likely a transient problem. "
-                        "Please, try again."
+                        "If it persists, please reach out to support@fal.ai with the following details: "  # noqa: E501
+                        f"{msg}"
                     )
-                elif e.details().endswith("died with <Signals.SIGKILL: 9>.`."):
+                elif msg.endswith("died with <Signals.SIGKILL: 9>.`."):
                     raise FalServerlessError(
                         "Isolated function crashed. "
                         "This is likely due to resource overflow. "
                         "You can try again by setting a bigger `machine_type`"
                     )
-
                 elif e.code() == grpc.StatusCode.INVALID_ARGUMENT and (
-                    "The function function could not be deserialized" in e.details()
+                    "The function function could not be deserialized" in msg
                 ):
-                    raise FalMissingDependencyError(e.details()) from None
+                    raise FalMissingDependencyError(msg) from None
                 else:
-                    raise FalServerlessError(e.details())
+                    raise FalServerlessError(msg)
 
         return handler
 
@@ -346,12 +396,18 @@ class FalServerlessHost(Host):
     _SUPPORTED_KEYS = frozenset(
         {
             "machine_type",
+            "machine_types",
+            "num_gpus",
             "keep_alive",
             "max_concurrency",
             "min_concurrency",
+            "concurrency_buffer",
             "max_multiplexing",
             "setup_function",
             "metadata",
+            "request_timeout",
+            "startup_timeout",
+            "private_logs",
             "_base_image",
             "_scheduler",
             "_scheduler_options",
@@ -364,8 +420,16 @@ class FalServerlessHost(Host):
 
     _log_printer = IsolateLogPrinter(debug=flags.DEBUG)
 
+    _thread_pool: ThreadPoolExecutor = field(default_factory=ThreadPoolExecutor)
+
+    def __getstate__(self) -> dict[str, Any]:
+        state = self.__dict__.copy()
+        state["_thread_pool"] = None
+        return state
+
     def __setstate__(self, state: dict[str, Any]) -> None:
         self.__dict__.update(state)
+        self._thread_pool = ThreadPoolExecutor()
         self.credentials = get_agent_credentials(self.credentials)
 
     @property
@@ -382,25 +446,32 @@ class FalServerlessHost(Host):
         application_name: str | None = None,
         application_auth_mode: Literal["public", "shared", "private"] | None = None,
         metadata: dict[str, Any] | None = None,
+        deployment_strategy: Literal["recreate", "rolling"] = "recreate",
+        scale: bool = True,
     ) -> str | None:
+        from isolate.backends.common import active_python
+
         environment_options = options.environment.copy()
         environment_options.setdefault("python_version", active_python())
         environments = [self._connection.define_environment(**environment_options)]
 
-        machine_type = options.host.get(
+        machine_type: list[str] | str = options.host.get(
             "machine_type", FAL_SERVERLESS_DEFAULT_MACHINE_TYPE
         )
         keep_alive = options.host.get("keep_alive", FAL_SERVERLESS_DEFAULT_KEEP_ALIVE)
-        max_concurrency = options.host.get("max_concurrency")
-        min_concurrency = options.host.get("min_concurrency")
-        max_multiplexing = options.host.get("max_multiplexing")
         base_image = options.host.get("_base_image", None)
         scheduler = options.host.get("_scheduler", None)
         scheduler_options = options.host.get("_scheduler_options", None)
+        max_concurrency = options.host.get("max_concurrency")
+        min_concurrency = options.host.get("min_concurrency")
+        concurrency_buffer = options.host.get("concurrency_buffer")
+        max_multiplexing = options.host.get("max_multiplexing")
         exposed_port = options.get_exposed_port()
-
+        request_timeout = options.host.get("request_timeout")
+        startup_timeout = options.host.get("startup_timeout")
         machine_requirements = MachineRequirements(
-            machine_type=machine_type,
+            machine_types=machine_type,  # type: ignore
+            num_gpus=options.host.get("num_gpus"),
             keep_alive=keep_alive,
             base_image=base_image,
             exposed_port=exposed_port,
@@ -409,6 +480,9 @@ class FalServerlessHost(Host):
             max_multiplexing=max_multiplexing,
             max_concurrency=max_concurrency,
             min_concurrency=min_concurrency,
+            concurrency_buffer=concurrency_buffer,
+            request_timeout=request_timeout,
+            startup_timeout=startup_timeout,
         )
 
         partial_func = _prepare_partial_func(func)
@@ -420,20 +494,19 @@ class FalServerlessHost(Host):
         if isinstance(func, ServeWrapper):
             # Assigning in a separate property leaving a place for the user
             # to add more metadata in the future
-            try:
-                metadata["openapi"] = func.openapi()
-            except Exception as e:
-                print(
-                    f"[warning] Failed to generate OpenAPI metadata for function: {e}"
-                )
+            metadata["openapi"] = func.openapi()
 
         for partial_result in self._connection.register(
             partial_func,
             environments,
             application_name=application_name,
-            application_auth_mode=application_auth_mode,
+            auth_mode=application_auth_mode,
             machine_requirements=machine_requirements,
             metadata=metadata,
+            deployment_strategy=deployment_strategy,
+            scale=scale,
+            # By default, logs are public
+            private_logs=options.host.get("private_logs", False),
         ):
             for log in partial_result.logs:
                 self._log_printer.print(log)
@@ -444,32 +517,38 @@ class FalServerlessHost(Host):
         return None
 
     @_handle_grpc_error()
-    def run(
+    def _run(
         self,
         func: Callable[..., ReturnT],
         options: Options,
         args: tuple[Any, ...],
         kwargs: dict[str, Any],
+        result_handler: Callable[..., None],
     ) -> ReturnT:
+        from isolate.backends.common import active_python
+
         environment_options = options.environment.copy()
         environment_options.setdefault("python_version", active_python())
         environments = [self._connection.define_environment(**environment_options)]
 
-        machine_type = options.host.get(
+        machine_type: list[str] | str = options.host.get(
             "machine_type", FAL_SERVERLESS_DEFAULT_MACHINE_TYPE
         )
         keep_alive = options.host.get("keep_alive", FAL_SERVERLESS_DEFAULT_KEEP_ALIVE)
         max_concurrency = options.host.get("max_concurrency")
         min_concurrency = options.host.get("min_concurrency")
+        concurrency_buffer = options.host.get("concurrency_buffer")
         max_multiplexing = options.host.get("max_multiplexing")
         base_image = options.host.get("_base_image", None)
         scheduler = options.host.get("_scheduler", None)
         scheduler_options = options.host.get("_scheduler_options", None)
         exposed_port = options.get_exposed_port()
         setup_function = options.host.get("setup_function", None)
-
+        request_timeout = options.host.get("request_timeout")
+        startup_timeout = options.host.get("startup_timeout")
         machine_requirements = MachineRequirements(
-            machine_type=machine_type,
+            machine_types=machine_type,  # type: ignore
+            num_gpus=options.host.get("num_gpus"),
             keep_alive=keep_alive,
             base_image=base_image,
             exposed_port=exposed_port,
@@ -478,6 +557,9 @@ class FalServerlessHost(Host):
             max_multiplexing=max_multiplexing,
             max_concurrency=max_concurrency,
             min_concurrency=min_concurrency,
+            concurrency_buffer=concurrency_buffer,
+            request_timeout=request_timeout,
+            startup_timeout=startup_timeout,
         )
 
         return_value = _UNSET
@@ -490,8 +572,7 @@ class FalServerlessHost(Host):
             machine_requirements=machine_requirements,
             setup_function=setup_function,
         ):
-            for log in partial_result.logs:
-                self._log_printer.print(log)
+            result_handler(partial_result)
 
             if partial_result.status.state is not HostedRunState.IN_PROGRESS:
                 state = partial_result.status.state
@@ -511,6 +592,46 @@ class FalServerlessHost(Host):
 
         return cast(ReturnT, return_value)
 
+    def run(
+        self,
+        func: Callable[..., ReturnT],
+        options: Options,
+        args: tuple[Any, ...],
+        kwargs: dict[str, Any],
+    ) -> ReturnT:
+        def result_handler(partial_result):
+            for log in partial_result.logs:
+                self._log_printer.print(log)
+
+        return self._run(func, options, args, kwargs, result_handler=result_handler)
+
+    def spawn(
+        self,
+        func: Callable[..., ReturnT],
+        options: Options,
+        args: tuple[Any, ...],
+        kwargs: dict[str, Any],
+    ) -> SpawnInfo:
+        ret = SpawnInfo()
+
+        def result_handler(partial_result):
+            ret.stream = partial_result.stream
+            for log in partial_result.logs:
+                if "And API access through" in log.message:
+                    ret.url = log.message.rsplit()[-1].replace("queue.", "")
+                ret.logs.put(log)
+
+        self._thread_pool.submit(
+            self._run,
+            func,
+            options,
+            args,
+            kwargs,
+            result_handler=result_handler,
+        )
+
+        return ret
+
 
 @dataclass
 class Options:
@@ -520,13 +641,14 @@ class Options:
 
     def add_requirements(self, requirements: list[str]):
         kind = self.environment["kind"]
-        if kind == "virtualenv":
+        if kind in ["virtualenv", "container"]:
             pip_requirements = self.environment.setdefault("requirements", [])
         elif kind == "conda":
             pip_requirements = self.environment.setdefault("pip", [])
         else:
             raise FalServerlessError(
-                "Only conda and virtualenv is supported as environment options"
+                "Only {conda, virtualenv, container} "
+                "are supported as environment options."
             )
 
         # Already has these.
@@ -542,7 +664,6 @@ class Options:
             return self.gateway.get("exposed_port")
 
 
-_DEFAULT_HOST = FalServerlessHost()
 _SERVE_PORT = 8080
 
 # Overload @function to help users identify the correct signature.
@@ -563,10 +684,10 @@ def function(
     serve: Literal[False] = False,
     exposed_port: int | None = None,
     max_concurrency: int | None = None,
+    local_python_modules: list[str] | None = None,
 ) -> Callable[
     [Callable[Concatenate[ArgsT], ReturnT]], IsolatedFunction[ArgsT, ReturnT]
-]:
-    ...
+]: ...
 
 
 @overload
@@ -580,10 +701,10 @@ def function(
     serve: Literal[True],
     exposed_port: int | None = None,
     max_concurrency: int | None = None,
+    local_python_modules: list[str] | None = None,
 ) -> Callable[
     [Callable[Concatenate[ArgsT], ReturnT]], ServedIsolatedFunction[ArgsT, ReturnT]
-]:
-    ...
+]: ...
 
 
 ### FalServerlessHost
@@ -594,23 +715,27 @@ def function(
     python_version: str | None = None,
     requirements: list[str] | None = None,
     # Common options
-    host: FalServerlessHost = _DEFAULT_HOST,
+    host: FalServerlessHost | None = None,
     serve: Literal[False] = False,
     exposed_port: int | None = None,
     max_concurrency: int | None = None,
+    local_python_modules: list[str] | None = None,
     # FalServerlessHost options
     metadata: dict[str, Any] | None = None,
-    machine_type: str = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
+    machine_type: str | list[str] = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
+    num_gpus: int | None = None,
     keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
     max_multiplexing: int = FAL_SERVERLESS_DEFAULT_MAX_MULTIPLEXING,
     min_concurrency: int = FAL_SERVERLESS_DEFAULT_MIN_CONCURRENCY,
+    concurrency_buffer: int = FAL_SERVERLESS_DEFAULT_CONCURRENCY_BUFFER,
+    request_timeout: int | None = None,
+    startup_timeout: int | None = None,
     setup_function: Callable[..., None] | None = None,
     _base_image: str | None = None,
     _scheduler: str | None = None,
 ) -> Callable[
     [Callable[Concatenate[ArgsT], ReturnT]], IsolatedFunction[ArgsT, ReturnT]
-]:
-    ...
+]: ...
 
 
 @overload
@@ -620,23 +745,27 @@ def function(
     python_version: str | None = None,
     requirements: list[str] | None = None,
     # Common options
-    host: FalServerlessHost = _DEFAULT_HOST,
+    host: FalServerlessHost | None = None,
     serve: Literal[True],
     exposed_port: int | None = None,
     max_concurrency: int | None = None,
+    local_python_modules: list[str] | None = None,
     # FalServerlessHost options
     metadata: dict[str, Any] | None = None,
-    machine_type: str = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
+    machine_type: str | list[str] = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
+    num_gpus: int | None = None,
     keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
     max_multiplexing: int = FAL_SERVERLESS_DEFAULT_MAX_MULTIPLEXING,
     min_concurrency: int = FAL_SERVERLESS_DEFAULT_MIN_CONCURRENCY,
+    concurrency_buffer: int = FAL_SERVERLESS_DEFAULT_CONCURRENCY_BUFFER,
+    request_timeout: int | None = None,
+    startup_timeout: int | None = None,
     setup_function: Callable[..., None] | None = None,
     _base_image: str | None = None,
     _scheduler: str | None = None,
 ) -> Callable[
     [Callable[Concatenate[ArgsT], ReturnT]], ServedIsolatedFunction[ArgsT, ReturnT]
-]:
-    ...
+]: ...
 
 
 ## conda
@@ -657,10 +786,10 @@ def function(
     serve: Literal[False] = False,
     exposed_port: int | None = None,
     max_concurrency: int | None = None,
+    local_python_modules: list[str] | None = None,
 ) -> Callable[
     [Callable[Concatenate[ArgsT], ReturnT]], IsolatedFunction[ArgsT, ReturnT]
-]:
-    ...
+]: ...
 
 
 @overload
@@ -679,10 +808,10 @@ def function(
     serve: Literal[True],
     exposed_port: int | None = None,
     max_concurrency: int | None = None,
+    local_python_modules: list[str] | None = None,
 ) -> Callable[
     [Callable[Concatenate[ArgsT], ReturnT]], ServedIsolatedFunction[ArgsT, ReturnT]
-]:
-    ...
+]: ...
 
 
 ### FalServerlessHost
@@ -698,23 +827,27 @@ def function(
     pip: list[str] | None = None,
     channels: list[str] | None = None,
     # Common options
-    host: FalServerlessHost = _DEFAULT_HOST,
+    host: FalServerlessHost | None = None,
     serve: Literal[False] = False,
     exposed_port: int | None = None,
     max_concurrency: int | None = None,
+    local_python_modules: list[str] | None = None,
     # FalServerlessHost options
     metadata: dict[str, Any] | None = None,
-    machine_type: str = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
+    machine_type: str | list[str] = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
+    num_gpus: int | None = None,
     keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
     max_multiplexing: int = FAL_SERVERLESS_DEFAULT_MAX_MULTIPLEXING,
     min_concurrency: int = FAL_SERVERLESS_DEFAULT_MIN_CONCURRENCY,
+    concurrency_buffer: int = FAL_SERVERLESS_DEFAULT_CONCURRENCY_BUFFER,
+    request_timeout: int | None = None,
+    startup_timeout: int | None = None,
     setup_function: Callable[..., None] | None = None,
     _base_image: str | None = None,
     _scheduler: str | None = None,
 ) -> Callable[
     [Callable[Concatenate[ArgsT], ReturnT]], IsolatedFunction[ArgsT, ReturnT]
-]:
-    ...
+]: ...
 
 
 @overload
@@ -729,38 +862,118 @@ def function(
     pip: list[str] | None = None,
     channels: list[str] | None = None,
     # Common options
-    host: FalServerlessHost = _DEFAULT_HOST,
+    host: FalServerlessHost | None = None,
     serve: Literal[True],
     exposed_port: int | None = None,
     max_concurrency: int | None = None,
+    local_python_modules: list[str] | None = None,
     # FalServerlessHost options
     metadata: dict[str, Any] | None = None,
-    machine_type: str = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
+    machine_type: str | list[str] = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
+    num_gpus: int | None = None,
     keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
     max_multiplexing: int = FAL_SERVERLESS_DEFAULT_MAX_MULTIPLEXING,
     min_concurrency: int = FAL_SERVERLESS_DEFAULT_MIN_CONCURRENCY,
+    concurrency_buffer: int = FAL_SERVERLESS_DEFAULT_CONCURRENCY_BUFFER,
+    request_timeout: int | None = None,
+    startup_timeout: int | None = None,
     setup_function: Callable[..., None] | None = None,
     _base_image: str | None = None,
     _scheduler: str | None = None,
 ) -> Callable[
     [Callable[Concatenate[ArgsT], ReturnT]], ServedIsolatedFunction[ArgsT, ReturnT]
-]:
-    ...
+]: ...
+
+
+@overload
+def function(
+    kind: Literal["container"],
+    *,
+    image: ContainerImage | None = None,
+    # Common options
+    host: FalServerlessHost | None = None,
+    serve: Literal[False] = False,
+    exposed_port: int | None = None,
+    max_concurrency: int | None = None,
+    local_python_modules: list[str] | None = None,
+    # FalServerlessHost options
+    metadata: dict[str, Any] | None = None,
+    machine_type: str | list[str] = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
+    num_gpus: int | None = None,
+    keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
+    max_multiplexing: int = FAL_SERVERLESS_DEFAULT_MAX_MULTIPLEXING,
+    min_concurrency: int = FAL_SERVERLESS_DEFAULT_MIN_CONCURRENCY,
+    concurrency_buffer: int = FAL_SERVERLESS_DEFAULT_CONCURRENCY_BUFFER,
+    request_timeout: int | None = None,
+    startup_timeout: int | None = None,
+    setup_function: Callable[..., None] | None = None,
+    _base_image: str | None = None,
+    _scheduler: str | None = None,
+) -> Callable[
+    [Callable[Concatenate[ArgsT], ReturnT]], IsolatedFunction[ArgsT, ReturnT]
+]: ...
+
+
+@overload
+def function(
+    kind: Literal["container"],
+    *,
+    image: ContainerImage | None = None,
+    # Common options
+    host: FalServerlessHost | None = None,
+    serve: Literal[True],
+    exposed_port: int | None = None,
+    max_concurrency: int | None = None,
+    local_python_modules: list[str] | None = None,
+    # FalServerlessHost options
+    metadata: dict[str, Any] | None = None,
+    machine_type: str | list[str] = FAL_SERVERLESS_DEFAULT_MACHINE_TYPE,
+    num_gpus: int | None = None,
+    keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE,
+    max_multiplexing: int = FAL_SERVERLESS_DEFAULT_MAX_MULTIPLEXING,
+    min_concurrency: int = FAL_SERVERLESS_DEFAULT_MIN_CONCURRENCY,
+    concurrency_buffer: int = FAL_SERVERLESS_DEFAULT_CONCURRENCY_BUFFER,
+    request_timeout: int | None = None,
+    startup_timeout: int | None = None,
+    setup_function: Callable[..., None] | None = None,
+    _base_image: str | None = None,
+    _scheduler: str | None = None,
+) -> Callable[
+    [Callable[Concatenate[ArgsT], ReturnT]], ServedIsolatedFunction[ArgsT, ReturnT]
+]: ...
 
 
 # implementation
 def function(  # type: ignore
     kind: str = "virtualenv",
     *,
-    host: Host = _DEFAULT_HOST,
+    host: Host | None = None,
+    local_python_modules: list[str] | None = None,
     **config: Any,
 ):
+    if host is None:
+        host = FalServerlessHost()
     options = host.parse_options(kind=kind, **config)
 
     def wrapper(func: Callable[ArgsT, ReturnT]):
         include_modules_from(func)
+
+        if local_python_modules and not isinstance(local_python_modules, list):
+            raise ValueError(
+                "local_python_modules must be a list of module names as strings, got "
+                f"{repr(local_python_modules)}"
+            )
+
+        for idx, module_name in enumerate(local_python_modules or []):
+            if not isinstance(module_name, str):
+                raise ValueError(
+                    "local_python_modules must be a list of module names as strings, "
+                    f"got {repr(module_name)} at index {idx}"
+                )
+            include_module(module_name)
+
         proxy = IsolatedFunction(
-            host=host,
+            host=host,  # type: ignore
             raw_func=func,  # type: ignore
             options=options,
         )
@@ -824,6 +1037,8 @@ class RouteSignature(NamedTuple):
 
 
 class BaseServable:
+    version: ClassVar[str] = "unknown"
+
     def collect_routes(self) -> dict[RouteSignature, Callable[..., Any]]:
         raise NotImplementedError
 
@@ -841,6 +1056,9 @@ class BaseServable:
         yield
 
     def _build_app(self) -> FastAPI:
+        import json
+        import traceback
+
         from fastapi import HTTPException, Request
         from fastapi.middleware.cors import CORSMiddleware
         from fastapi.responses import JSONResponse
@@ -881,6 +1099,60 @@ class BaseServable:
                 # If it's not a generic 404, just return the original message.
                 return JSONResponse({"detail": exc.detail}, 404)
 
+        @_app.exception_handler(AppException)
+        async def app_exception_handler(request: Request, exc: AppException):
+            return JSONResponse({"detail": exc.message}, exc.status_code)
+
+        @_app.exception_handler(FieldException)
+        async def field_exception_handler(request: Request, exc: FieldException):
+            return JSONResponse(exc.to_pydantic_format(), exc.status_code)
+
+        # ref: https://github.com/fastapi/fastapi/blob/37c8e7d76b4b47eb2c4cced6b4de59eb3d5f08eb/fastapi/exception_handlers.py#L20
+        @_app.exception_handler(RequestValidationError)
+        async def request_val_exception_handler(
+            request: Request, exc: RequestValidationError
+        ):
+            return JSONResponse(
+                {"detail": jsonable_encoder(exc.errors())},
+                422,
+                headers={"x-fal-billable-units": "0"},
+            )
+
+        @_app.exception_handler(CUDAOutOfMemoryException)
+        async def cuda_out_of_memory_exception_handler(
+            request: Request, exc: CUDAOutOfMemoryException
+        ):
+            return JSONResponse({"detail": exc.message}, exc.status_code)
+
+        @_app.exception_handler(Exception)
+        async def traceback_logging_exception_handler(request: Request, exc: Exception):
+            _, MINOR, *_ = sys.version_info
+
+            # traceback.format_exception() has a different signature in Python >=3.10
+            if MINOR >= 10:
+                formatted_exception = traceback.format_exception(exc)  # type: ignore
+            else:
+                formatted_exception = traceback.format_exception(
+                    type(exc), exc, exc.__traceback__
+                )
+
+            print(
+                json.dumps({"traceback": "".join(formatted_exception[::-1])}),
+                flush=True,
+            )
+
+            if _is_cuda_oom_exception(exc):
+                return await cuda_out_of_memory_exception_handler(
+                    request, CUDAOutOfMemoryException()
+                )
+
+            # last line of defense against misc GPU errors that could indicate a bad
+            # worker
+            if any(marker in str(exc).lower() for marker in ["cuda", "cudnn", "nvml"]):
+                return JSONResponse({"detail": "GPU error"}, 503)
+
+            return JSONResponse({"detail": "Internal Server Error"}, 500)
+
         routes = self.collect_routes()
         if not routes:
             raise ValueError("An application must have at least one route!")
@@ -909,16 +1181,31 @@ class BaseServable:
         Build the OpenAPI specification for the served function.
         Attach needed metadata for a better integration to fal.
         """
-        return self._build_app().openapi()
+        try:
+            return self._build_app().openapi()
+        except Exception as e:
+            raise FalServerlessException(
+                "Failed to generate OpenAPI metadata for function"
+            ) from e
 
     def serve(self) -> None:
         import asyncio
 
+        from prometheus_client import Gauge
         from starlette_exporter import handle_metrics
         from uvicorn import Config
 
+        # NOTE: this uses the global prometheus registry
+        app_info = Gauge("fal_app_info", "Fal application information", ["version"])
+        app_info.labels(version=self.version).set(1)
+
         app = self._build_app()
-        server = Server(config=Config(app, host="0.0.0.0", port=8080))
+
+        # We use the default workers=1 config because setup function can be heavy
+        # and it runs once per worker.
+        server = Server(
+            config=Config(app, host="0.0.0.0", port=8080, timeout_keep_alive=300)
+        )
         metrics_app = FastAPI()
         metrics_app.add_route("/metrics", handle_metrics)
         metrics_server = Server(config=Config(metrics_app, host="0.0.0.0", port=9090))
@@ -930,7 +1217,8 @@ class BaseServable:
             }
 
             _, pending = await asyncio.wait(
-                tasks.keys(), return_when=asyncio.FIRST_COMPLETED,
+                tasks.keys(),
+                return_when=asyncio.FIRST_COMPLETED,
             )
             if not pending:
                 return
@@ -1004,6 +1292,14 @@ class IsolatedFunction(Generic[ArgsT, ReturnT]):
         )
         return future
 
+    def spawn(self, *args: ArgsT.args, **kwargs: ArgsT.kwargs):
+        return self.host.spawn(
+            self.func,
+            self.options,
+            args,
+            kwargs,
+        )
+
     def __call__(self, *args: ArgsT.args, **kwargs: ArgsT.kwargs) -> ReturnT:
         try:
             return self.host.run(
@@ -1041,14 +1337,12 @@ class IsolatedFunction(Generic[ArgsT, ReturnT]):
     @overload
     def on(
         self, host: Host | None = None, *, serve: Literal[False] = False, **config: Any
-    ) -> IsolatedFunction[ArgsT, ReturnT]:
-        ...
+    ) -> IsolatedFunction[ArgsT, ReturnT]: ...
 
     @overload
     def on(
         self, host: Host | None = None, *, serve: Literal[True], **config: Any
-    ) -> ServedIsolatedFunction[ArgsT, ReturnT]:
-        ...
+    ) -> ServedIsolatedFunction[ArgsT, ReturnT]: ...
 
     def on(self, host: Host | None = None, **config: Any):  # type: ignore
         host = host or self.host
@@ -1102,14 +1396,12 @@ class ServedIsolatedFunction(
     @overload  # type: ignore[override,no-overload-impl]
     def on(  # type: ignore[no-overload-impl]
         self, host: Host | None = None, *, serve: Literal[True] = True, **config: Any
-    ) -> ServedIsolatedFunction[ArgsT, ReturnT]:
-        ...
+    ) -> ServedIsolatedFunction[ArgsT, ReturnT]: ...
 
     @overload
     def on(
         self, host: Host | None = None, *, serve: Literal[False], **config: Any
-    ) -> IsolatedFunction[ArgsT, ReturnT]:
-        ...
+    ) -> IsolatedFunction[ArgsT, ReturnT]: ...
 
 
 class Server(uvicorn.Server):
@@ -1122,4 +1414,3 @@ class Server(uvicorn.Server):
 
     def install_signal_handlers(self) -> None:
         pass
-
