@@ -31,18 +31,28 @@ def test_set_rev():
 def test_scale():
     args = parse_args(
         [
-            "apps", "scale",
-            "--keep-alive", "123",
-            "--max-multiplexing", "321",
-            "--min-concurrency", "7",
-            "--max-concurrency", "10",
+            "apps",
+            "scale",
+            "myapp",
+            "--keep-alive",
+            "123",
+            "--max-multiplexing",
+            "321",
+            "--min-concurrency",
+            "7",
+            "--max-concurrency",
+            "10",
+            "--concurrency-buffer",
+            "3",
         ]
     )
     assert args.func == _scale
+    assert args.app_name == "myapp"
     assert args.keep_alive == 123
     assert args.max_multiplexing == 321
     assert args.min_concurrency == 7
     assert args.max_concurrency == 10
+    assert args.concurrency_buffer == 3
 
 
 def test_runners():
@@ -60,4 +70,3 @@ def test_delete_rev():
     args = parse_args(["apps", "delete-rev", "myrev"])
     assert args.func == _delete_rev
     assert args.app_rev == "myrev"
-
