@@ -1,16 +1,30 @@
 from __future__ import annotations
 
+import subprocess
+import uuid
 from contextlib import suppress
 
-import fal
 import pytest
-from fal.api import FalServerlessError, Options
-from fal.toolkit.file.file import File
+from isolate.backends.common import active_python
 from pydantic import __version__ as pydantic_version
 
-PACKAGE_NAME = "fal"
+import fal
+from fal.api import FalServerlessError, Options
+from fal.container import ContainerImage
+from fal.toolkit.file import File
+
+PACKAGE_NAME = "fall"
 
 
+def git_revision_short_hash() -> str:
+    return (
+        subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
+        .decode("ascii")
+        .strip()
+    )
+
+
+@pytest.mark.xfail(reason="Temporary mismatch due to grpc version updates. Ping @efiop")
 def test_missing_dependencies_nested_server_error(isolated_client):
     @isolated_client()
     def test1():
@@ -55,6 +69,88 @@ def test_regular_function_on_nomad(isolated_client):
     assert mult(5, 2) == 10
 
 
+def test_regular_function_in_a_container(isolated_client):
+    @isolated_client("container")
+    def regular_function():
+        return 42
+
+    assert regular_function() == 42
+
+    @isolated_client("container")
+    def mult(a, b):
+        return a * b
+
+    assert mult(5, 2) == 10
+
+
+def test_container_no_venv(isolated_client):
+    actual_python = active_python()
+
+    @isolated_client(
+        "container",
+        image=ContainerImage.from_dockerfile_str(
+            f"""
+            FROM python:{actual_python}
+            # {uuid.uuid4()}
+            """
+        ),
+    )
+    def myfunc():
+        return 42
+
+    assert myfunc() == 42
+
+
+def test_container_venv(isolated_client):
+    actual_python = active_python()
+
+    @isolated_client(
+        "container",
+        image=ContainerImage.from_dockerfile_str(
+            f"""
+            FROM python:{actual_python}
+            # {uuid.uuid4()}
+            RUN python -m venv .venv
+            ENV PATH="/root/.venv/bin:$PATH"
+            """
+        ),
+    )
+    def myfunc():
+        return 42
+
+    assert myfunc() == 42
+
+
+def test_regular_function_in_a_container_with_custom_image(isolated_client):
+    actual_python = active_python()
+
+    @isolated_client(
+        "container",
+        image=ContainerImage.from_dockerfile_str(
+            f"""
+            FROM python:{actual_python}-slim
+            # {git_revision_short_hash()}
+            RUN env
+            """
+        ),
+    )
+    def regular_function():
+        return 42
+
+    assert regular_function() == 42
+
+    @isolated_client(
+        "container",
+        image=ContainerImage.from_dockerfile_str(
+            f"FROM python:{actual_python}-slim\n# {git_revision_short_hash()}"
+        ),
+    )
+    def mult(a, b):
+        return a * b
+
+    assert mult(5, 2) == 10
+
+
 def test_function_pipelining(isolated_client):
     @isolated_client("virtualenv")
     def regular_function():
@@ -67,6 +163,40 @@ def test_function_pipelining(isolated_client):
     assert calling_function(regular_function()) == 84
 
 
+def test_function_exception(isolated_client):
+    class FooException(Exception):
+        pass
+
+    class BarException(Exception):
+        pass
+
+    class MyFuncException(Exception):
+        pass
+
+    def foo():
+        raise FooException("foo")
+
+    def bar():
+        try:
+            foo()
+        except Exception as exc:
+            raise BarException("bar") from exc
+
+    @isolated_client()
+    def myfunc():
+        try:
+            bar()
+        except Exception as exc:
+            raise MyFuncException("myfunc") from exc
+
+    with pytest.raises(MyFuncException) as excinfo:
+        myfunc()
+
+    bar_exc = excinfo.value.__cause__
+    assert isinstance(bar_exc, BarException)
+    assert isinstance(bar_exc.__cause__, FooException)
+
+
 @pytest.mark.xfail(reason="See https://github.com/fal-ai/fal/issues/169")
 def test_function_calling_other_function(isolated_client):
     try:
@@ -341,22 +471,6 @@ def test_futures(isolated_client):
     assert future_4.result() == 8
 
 
-def test_conda_environment(isolated_client):
-    @isolated_client(
-        "conda",
-        packages=["pyjokes=0.6.0"],
-        machine_type="L",
-        # conda is only supported on Kubernetes
-        _scheduler="kubernetes",
-    )
-    def regular_function():
-        import pyjokes
-
-        return pyjokes.__version__
-
-    assert regular_function() == "0.6.0"
-
-
 @pytest.mark.xfail(reason="Nomad does not support conda")
 def test_conda_environment_on_nomad(isolated_client):
     @isolated_client(
@@ -374,6 +488,7 @@ def test_conda_environment_on_nomad(isolated_client):
     assert regular_function() == "0.6.0"
 
 
+@pytest.mark.xfail(reason="Broken on nomad")
 def test_cached_function(isolated_client, capsys, monkeypatch):
     import inspect
     import time
@@ -410,7 +525,6 @@ def test_cached_function(isolated_client, capsys, monkeypatch):
         "virtualenv",
         keep_alive=30,
         requirements=["pyjokes     "],
-        _scheduler="kubernetes",
     )
     def regular_function(n):
         if get_pipe() == "pipe":
@@ -484,6 +598,8 @@ def test_serve_on_off(isolated_client):
 
 
 def test_worker_env_vars(isolated_client):
+    from fal.flags import GRPC_HOST
+
     @isolated_client("virtualenv", keep_alive=5)
     def get_env_var(name: str) -> str | None:
         import os
@@ -492,8 +608,8 @@ def test_worker_env_vars(isolated_client):
 
     fal_host = get_env_var("FAL_HOST")
     assert fal_host, "FAL_HOST is not set"
-    assert fal_host.startswith("api.")
-    assert fal_host.endswith(".shark.fal.ai")
+    assert fal_host == GRPC_HOST, "FAL_HOST is not set to the correct value"
+    print(f"FAL_HOST: {fal_host}, GRPC_HOST: {GRPC_HOST}")
 
     fal_key_id = get_env_var("FAL_KEY_ID")
     assert fal_key_id, "FAL_KEY_ID is not set"
@@ -502,23 +618,45 @@ def test_worker_env_vars(isolated_client):
     assert fal_key_secret, "FAL_KEY_SECRET is not set"
 
 
-def test_fal_storage(isolated_client):
-    file = File.from_bytes(b"Hello fal storage from local", repository="fal")
-    assert file.url.startswith(
-        "https://storage.googleapis.com/isolate-dev-smiling-shark_toolkit_bucket/"
+@pytest.mark.parametrize(
+    "repo_type, url_prefix",
+    [
+        (
+            "fal",
+            "https://storage.googleapis.com/isolate-dev-smiling-shark_toolkit_public_bucket/",
+        ),
+        ("fal_v2", "https://v2.fal.media/files"),
+        ("fal_v3", "https://v3.fal.media/files"),
+    ],
+)
+def test_fal_storage(isolated_client, repo_type, url_prefix):
+    file = File.from_bytes(
+        b"Hello fal storage from local", repository=repo_type, fallback_repository=None
     )
+    assert file.url.startswith(url_prefix)
     assert file.as_bytes().decode().endswith("local")
 
-    @isolated_client(serve=True, requirements=[f"pydantic=={pydantic_version}"])
+    @isolated_client(
+        serve=True, requirements=[f"pydantic=={pydantic_version}", "tomli"]
+    )
     def hello_file():
         # Run in the isolated environment
-        return File.from_bytes(b"Hello fal storage from isolated", repository="fal")
+        return File.from_bytes(
+            b"Hello fal storage from isolated",
+            repository=repo_type,
+            save_kwargs={
+                "multipart": False,
+                "multipart_threshold": 1024 * 1024,
+                "multipart_chunk_size": 1024 * 1024,
+                "multipart_max_concurrency": 10,
+                "object_lifecycle_preference": {},
+            },
+            fallback_repository=None,
+        )
 
     local_fn = hello_file.on(serve=False)
     file = local_fn()
-    assert file.url.startswith(
-        "https://storage.googleapis.com/isolate-dev-smiling-shark_toolkit_bucket/"
-    )
+    assert file.url.startswith(url_prefix)
     assert file.as_bytes().decode().endswith("isolated")
 
 
